major, minor, patch = (0, 1, 1)
__version__ = f"{major}.{minor}.{patch}"
__author__ = "Unai Saralegui"
__email__ = "usaralegui@gmail.com"
__credits__ = ["Unai Saralegui"]
__maintainer__ = "Unai Saralegui"
__status__ = "Development"



import matplotlib.pyplot as plt
import networkx as nx
import numpy as np

from . import data
from . import utils


def floorplan_to_color(data: data.RplanData):
    """
    Get an image like object with the color for each pixel

    :param data: a data object from RPLAN dataset
    :type data: data.RplanData
    :return: list of colors for each pixel
    :rtype: list
    """
    coloured_image = [[utils.ROOM_COLOR[pix] for pix in row] for row in data.category]
    return coloured_image


def plot_floorplan(data: data.RplanData, ax=None, title=None):
    """
    Plot a floorplan

    :param data: a data object from RPLAN dataset
    :type data: data.RplanData
    :param ax: optional axes to plot in
    :type ax: matplotlib.axes._subplots.AxesSubplot
    :param title: optional title to add to the plot
    :type title: str
    :return: the plotted axes
    :rtype: matplotlib.axes._subplots.AxesSubplot
    """
    coloured_image = floorplan_to_color(data)
    if ax is None:
        ax = plt.subplot()
    ax.imshow(coloured_image)
    ax.set_title(title)
    ax.axis("off")
    return ax


def plot_floorplan_graph(data: data.RplanData, ax=None, title=None,
                         with_colors=True, edge_label=None):
    """
    Plot the graph representation for a floorplan

    :param data: a data object from RPLAN dataset
    :type data: data.RplanData
    :param ax: optional axes to plot in
    :type ax: matplotlib.axes._subplots.AxesSubplot
    :param title: optional title to add to the plot
    :type title: str
    :param with_colors: optional, whether to colour nodes by their class or not, defaults to True
    :type with_colors: bool
    :return: the plotted axes
    :rtype: matplotlib.axes._subplots.AxesSubplot
    """
    if ax is None:
        ax = plt.subplot()
    G = data.get_graph()
    pos = nx.spring_layout(G)
    if with_colors:
        colors = [np.array(utils.ROOM_COLOR.get(G.nodes[n]['category'], [255, 255, 255]))/255 for n in G.nodes]
        nx.draw(G, pos, with_labels=True, ax=ax, node_color=colors)
    else:
        nx.draw(G, pos, with_labels=True, ax=ax, )
    if edge_label:
        edge_labels = nx.get_edge_attributes(G, edge_label)
        nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels, ax=ax)
    ax.set_title(title)
    return ax



import setuptools
from distutils.util import convert_path

with open("README.md", "r") as fh:
    long_description = fh.read()

metadata = {}
ver_path = convert_path('rplanpy/metadata.py')
with open(ver_path) as ver_file:
    exec(ver_file.read(), metadata)


setuptools.setup(
    name="rplanpy",
    version=metadata['__version__'],
    author=metadata['__author__'],
    author_email=metadata['__email__'],
    description="Set of python utilities to work with the RPLAN dataset from \"[Data-driven Interior Plan Generation for Residential Buildings](https://doi.org/10.1145/3355089.3356556)\" paper.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/unaisaralegui/rplanpy",
    packages=setuptools.find_packages(),
    install_requires=[
        'imageio',
        'numpy',
        'matplotlib',
        'networkx',
        'scikit-image',
        'scipy'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



ROOM_CLASS = {
    0: "Living room",
    1: "Master room",
    2: "Kitchen",
    3: "Bathroom",
    4: "Dining room",
    5: "Child room",
    6: "Study room",
    7: "Second room",
    8: "Guest room",
    9: "Balcony",
    10: "Entrance",
    11: "Storage",
    12: "Wall-in",
    13: "External area",
    14: "Exterior wall",
    15: "Front door",
    16: "Interior wall",
    17: "Interior door",
}

ROOM_TYPE = {
    0: 'PublicArea',
    1: 'Bedroom',
    2: 'FunctionArea',
    3: 'FunctionArea',
    4: 'FunctionArea',
    5: 'Bedroom',
    6: 'Bedroom',
    7: 'Bedroom',
    8: 'Bedroom',
    9: 'PublicArea',
    10: 'PublicArea',
    11: 'PublicArea',
    12: 'PublicArea',
    13: 'External',
    14: 'ExteriorWall',
    15: 'FrontDoor',
    16: 'InteriorWall',
    17: 'InteriorDoor',
}

ROOM_COLOR = {
    0: [244, 242, 229],
    1: [253, 244, 171],
    2: [234, 216, 214],
    3: [205, 233, 252],
    4: [244, 242, 229],
    5: [253, 244, 171],
    6: [253, 244, 171],
    7: [253, 244, 171],
    8: [253, 244, 171],
    9: [208, 216, 135],
    10: [244, 242, 229],
    11: [249, 222, 189],
    12: [128, 128, 128],
    13: [255, 255, 255],
    14: [79, 79, 79],
    15: [255, 225, 25],
    16: [128, 128, 128],
    17: [255, 225, 25],
}


def get_image_channel(image, n):
    """
    Get channel n of an image

    :param image: An image loaded with imageio
    :param n: channel to get
    :return: values in specified channel
    """
    return image[..., n]


def collide2d(bbox1, bbox2, th=0):
    """
    Do two boxes collide with each other?
    :param bbox1: indexes of box 1 (y0, x0, y1, x1)
    :param bbox2: indexes of box 2 (y0, x0, y1, x1)
    :param th: optional margin to add to the boxes (default 0)
    :return: True/False depending if boxes do collide
    :rtype: bool
    """
    return not(
        (bbox1[0]-th > bbox2[2]) or
        (bbox1[2]+th < bbox2[0]) or
        (bbox1[1]-th > bbox2[3]) or
        (bbox1[3]+th < bbox2[1])
    )


def point_box_relation(u, vbox):
    """
    Check in which point is located related to a box
    :param u: point to check (y, x)
    :param vbox: box to check point with (y0, x0, y1, x1)
    :return: code with the location of the point
     0   3   8
        ---
     2 | 4 | 7
        ---
     1   6   9
    """
    uy, ux = u
    vy0, vx0, vy1, vx1 = vbox
    if (ux < vx0 and uy <= vy0) or (ux == vx0 and uy == vy0):
        relation = 0  # 'left-above'
    elif vx0 <= ux < vx1 and uy <= vy0:
        relation = 3  # 'above'
    elif (vx1 <= ux and uy < vy0) or (ux == vx1 and uy == vy0):
        relation = 8  # 'right-above'
    elif vx1 <= ux and vy0 <= uy < vy1:
        relation = 7  # 'right-of'
    elif (vx1 < ux and vy1 <= uy) or (ux == vx1 and uy == vy1):
        relation = 9  # 'right-below'
    elif vx0 < ux <= vx1 and vy1 <= uy:
        relation = 6  # 'below'
    elif (ux <= vx0 and vy1 < uy) or (ux == vx0 and uy == vy1):
        relation = 1  # 'left-below'
    elif ux <= vx0 and vy0 < uy <= vy1:
        relation = 2  # 'left-of'
    elif vx0 < ux < vx1 and vy0 < uy < vy1:
        relation = 4  # 'inside'
    else:
        relation = None
    return relation


def door_room_relation(door_box, room_box):
    y0, x0, y1, x1 = room_box[:4]
    yc, xc = (y1 + y0) / 2, (x0 + x1) / 2
    y0b, x0b, y1b, x1b = door_box[:4]
    y, x = (y1b + y0b) / 2, (x0b + x1b) / 2

    if x == xc and y < yc:
        return 7
    elif x == xc and y > yc:
        return 1
    elif y == yc and x < xc:
        return 10
    elif y == yc and x > xc:
        return 4
    elif x0 < x < xc:
        if y < yc:
            return 8
        else:
            return 12
    elif xc < x < x1:
        if y < yc:
            return 6
        else:
            return 2
    elif y0 < y < yc:
        if x < xc:
            return 9
        else:
            return 5
    elif yc < y < y1:
        if x < xc:
            return 11
        else:
            return 3
    else:
        return 0



import imageio
import numpy as np
import networkx as nx
from skimage import measure, feature, segmentation
from scipy import stats, ndimage


from . import utils


class RplanData:
    """
    Set of utilities for RPLAN dataset images

    :param image_path: Path to the image from the RPLAN dataset
    :type image_path: str
    """

    def __init__(self, image_path):
        self.image_path = image_path
        self.image = imageio.imread(image_path)
        self.boundary = utils.get_image_channel(self.image, 0)
        self.category = utils.get_image_channel(self.image, 1)
        self.instance = utils.get_image_channel(self.image, 2)
        self.inside = utils.get_image_channel(self.image, 3)
        self._rooms = None
        self._edges = None
        self._interior_doors = None

    def get_front_door_mask(self) -> np.array:
        """
        Get the mask for the front door

        :returns: Bounding box (min_row, min_col, max_row, max_col). Pixels belonging to the bounding box for the front door
        :rtype: np.ndarray
        """
        front_door_mask = self.boundary == 255
        region = measure.regionprops(front_door_mask.astype(int))[0]
        return np.array(region.bbox, dtype=int)

    def get_rooms(self) -> np.array:
        """
        Get the rooms in the floor plan
        :return:
        """
        if self._rooms is not None:
            return self._rooms
        rooms = []
        regions = measure.regionprops(self.instance)
        for region in regions:
            # c is the most common value in the region,
            # in this case c is the category of the room (channel 2)
            c = stats.mode(
                self.category[region.coords[:, 0],
                              region.coords[:, 1]],
                axis=None
            )[0][0]
            i = stats.mode(
                self.instance[region.coords[:, 0],
                              region.coords[:, 1]],
                axis=None
            )[0][0]
            y0, x0, y1, x1 = np.array(region.bbox)
            rooms.append([y0, x0, y1, x1, c, i])
        self._rooms = np.array(rooms, dtype=int)
        return self._rooms

    def get_edges(self, th: int = 9) -> np.array:
        if self._edges is not None:
            return self._edges
        if self._rooms is None:
            self.get_rooms()
        edges = []
        for u in range(len(self._rooms)):
            for v in range(u + 1, len(self._rooms)):
                # check if two rooms are connected, if they are check the spatial relation between them
                if not utils.collide2d(self._rooms[u, :4], self._rooms[v, :4], th=th):
                    # rooms are not connected
                    continue
                uy0, ux0, uy1, ux1, c1, i1 = self._rooms[u]
                vy0, vx0, vy1, vx1, c2, i2 = self._rooms[v]
                uc = (uy0 + uy1) / 2, (ux0 + ux1) / 2
                vc = (vy0 + vy1) / 2, (vx0 + vx1) / 2
                if ux0 < vx0 and ux1 > vx1 and uy0 < vy0 and uy1 > vy1:
                    relation = 5  # 'surrounding'
                elif ux0 >= vx0 and ux1 <= vx1 and uy0 >= vy0 and uy1 <= vy1:
                    relation = 4  # 'inside'
                else:
                    relation = utils.point_box_relation(uc, self._rooms[v, :4])
                edges.append([u, v, relation])
        self._edges = np.array(edges, dtype=int)
        return self._edges

    def get_interior_doors(self) -> np.array:
        if self._interior_doors is not None:
            return self._interior_doors
        doors = []
        category = 17  # InteriorDoor
        mask = (self.category == category).astype(np.uint8)
        distance = ndimage.morphology.distance_transform_cdt(mask)
        local_maxi = (distance > 1).astype(np.uint8)
        corner_measurement = feature.corner_harris(local_maxi)
        local_maxi[corner_measurement > 0] = 0
        markers = measure.label(local_maxi)

        labels = segmentation.watershed(-distance, markers, mask=mask, connectivity=8)
        regions = measure.regionprops(labels)

        for region in regions:
            y0, x0, y1, x1 = np.array(region.bbox)
            doors.append([y0, x0, y1, x1, category])

        self._interior_doors = np.array(doors, dtype=int)
        return self._interior_doors

    def get_rooms_with_properties(self) -> dict:
        """
        Get each room with their properties formatted in a dict

        :return: dictionary with room id as key and its category and additional properties as values
        :rtype: dict
        """
        properties_per_room = {
            room[-1]: {
                'category': room[-2],
                'bounding_box': room[0:4],
                'area': (room[2]-room[0]) * (room[3]-room[1])
            }
            for room in self.get_rooms()
        }
        return properties_per_room

    def door_in_edge(self, edge: list) -> bool:
        """
        Check if there is a door in an edge

        :param edge: edge to check [room1, room2, relation]
        :type edge: list
        :return: door in an edge?
        :rtype: bool
        """
        doors = self.get_interior_doors()
        room1 = self.get_rooms()[edge[0]]
        room2 = self.get_rooms()[edge[1]]
        for i in range(len(doors)):
            if utils.door_room_relation(doors[i], room1) and utils.door_room_relation(doors[i], room2):
                return True
        return False

    def get_edges_with_properties(self) -> list:
        """
        Get edges between rooms with their properties formatted in a dict

        :return: list of edges and edge properties [(u, v, props)]
        :rtype: list
        """
        rooms = self.get_rooms()
        properties_per_edge = [
            (
                rooms[edge[0]][-1], rooms[edge[1]][-1],
                {
                    'location': edge[2],
                    'door': self.door_in_edge(edge)
                }
            )
            for edge in self.get_edges()
        ]
        return properties_per_edge

    def get_graph(self) -> nx.classes.graph.Graph:
        """
        Get the graph representation for the floorplan

        :return: a networkx graph with the nodes, the edges and their properties
        :rtype: networkx.classes.graph.Graph
        """
        G = nx.Graph()
        # add nodes
        G.add_nodes_from([(room, props) for room, props in self.get_rooms_with_properties().items()])
        # add edges
        G.add_edges_from(self.get_edges_with_properties())
        return G
